"""
Benchmark functions for :mod:`locan.process.aggregate`
"""

import numpy as np

import locan as lc
from locan.process.aggregate import (
    _histogram_boost_histogram,
    _histogram_fast_histogram,
    _histogram_mean_boost_histogram,
    _histogram_mean_fast_histogram,
)


class BenchmarkBins:
    """
    Benchmarks for initializing Bins objects
    """

    def setup(self):
        self.bins = lc.Bins(n_bins=100, bin_range=(0, 1))

    def time_initialize_bins(self):
        """Time to initialize bins"""
        lc.Bins(n_bins=100, bin_range=(0, 1))

    def mem_bins(self):
        return self.bins


class BenchmarkHistogram:
    """
    Benchmarks for computing histograms
    """

    def setup(self):
        n_points = 1000
        self.points = np.vstack(
            [np.linspace(0, 1, n_points), np.linspace(0, 1, n_points)]
        )
        self.values = np.linspace(0, 1, n_points)
        self.bins = lc.Bins(n_bins=(100, 100), bin_range=(0, 1))

    def time__histogram_fast_histogram(self):
        _histogram_fast_histogram(data=self.points, bins=self.bins)

    def time__histogram_boost_histogram(self):
        _histogram_boost_histogram(data=self.points, bins=self.bins)

    def time__histogram_mean_fast_histogram(self):
        _histogram_mean_fast_histogram(
            data=self.points, bins=self.bins, values=self.values
        )

    def time__histogram_mean_boost_histogram(self):
        _histogram_mean_boost_histogram(
            data=self.points, bins=self.bins, values=self.values
        )


def main():
    bm = BenchmarkHistogram()
    bm.setup()
    bm.time__histogram_fast_histogram()
    bm.time__histogram_boost_histogram()
    bm.time__histogram_mean_fast_histogram()
    bm.time__histogram_mean_boost_histogram()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkHistogram()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time__histogram_fast_histogram()
        bm.time__histogram_boost_histogram()
        bm.time__histogram_mean_fast_histogram()
        bm.time__histogram_mean_boost_histogram()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    # main()
    main_profile()



"""

Render localization data with matplotlib.

This module provides functions to render and present localization data
making use of matplotlib.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   render2d
   render3d

"""

from __future__ import annotations

from locan.visualize.render_mpl import render2d, render3d

from .render2d import *
from .render3d import *

__all__: list[str] = []
__all__.extend(render2d.__all__)  # type: ignore
__all__.extend(render3d.__all__)  # type: ignore



from copy import copy

import google.protobuf.message
import pytest

from locan import (
    load_metadata_from_toml,
    merge_metadata,
    message_scheme,
    metadata_from_toml_string,
    metadata_to_formatted_string,
)
from locan.data import metadata_pb2
from locan.data.metadata_utils import _dict_to_protobuf, _modify_meta


@pytest.fixture
def metadata():
    metadata = metadata_pb2.Metadata()
    optical_unit = metadata.experiment.setups.add().optical_units.add()
    optical_unit.detection.camera.integration_time.FromMilliseconds(20)

    metadata.identifier = "111"
    metadata.comment = "comment"
    metadata.creation_time.FromJsonString("1111-11-11T11:11:11+01:00")

    return metadata


@pytest.fixture
def meta_dict():
    meta_dict_ = {
        "identifier": "123",
        "comment": "my comment",
        "ancestor_identifiers": ["1", "2"],
        "production_time": "2022-05-14T06:58:00Z",
        "localizer": {"software": "rapidSTORM"},
        "relations": [{"identifier": "1"}],
        "experiment": {
            "setups": [
                {
                    "identifier": "1",
                    "optical_units": [
                        {
                            "identifier": "1",
                            "detection": {
                                "camera": {
                                    "identifier": "1",
                                    "integration_time": 10000000,
                                }
                            },
                        }
                    ],
                }
            ]
        },
    }
    return meta_dict_


@pytest.fixture
def metadata_toml():
    metadata_toml_ = """
    # Define the class (message) instances.

    [[messages]]
    name = "metadata"
    module = "locan.data.metadata_pb2"
    class_name = "Metadata"


    # Fill metadata attributes
    # Use [[]] to add repeated elements
    # Use string '2022-05-14T06:58:00Z' for Timestamp elements
    # Use int in nanoseconds for Duration elements

    [metadata]
    identifier = "123"
    comment = "my comment"
    ancestor_identifiers = ["1", "2"]
    production_time = '2022-05-14T06:58:00Z'

    [metadata.localizer]
    software = "rapidSTORM"

    [[metadata.relations]]
    identifier = "1"

    [[metadata.experiment.setups]]
    identifier = "1"

    [[metadata.experiment.setups.optical_units]]
    identifier = "1"

    [metadata.experiment.setups.optical_units.detection.camera]
    identifier = "1"
    integration_time = 10_000_000
    """
    return metadata_toml_


def test__modify_meta(metadata, locdata_2d):
    new_locdata = copy(locdata_2d)
    new_metadata = _modify_meta(
        locdata_2d,
        new_locdata,
        function_name="test_function",
        parameter=dict(first=1, second=2),
        meta=None,
    )

    assert (
        metadata_to_formatted_string(message=new_metadata.history[-1], as_one_line=True)
        == "name: \"test_function\" parameter: \"{\\'first\\': 1, \\'second\\': 2}\""
    )


def test_metadata_to_formatted_string(metadata):
    results_string = metadata_to_formatted_string(message=metadata, as_one_line=True)
    results_string_expected = (
        'identifier: "111" comment: "comment" experiment { setups { optical_units { detection '
        "{ camera { integration_time { 0.020s } } } } } } creation_time { 1111-11-11T10:11:11Z }"
    )
    assert results_string == results_string_expected


def test__dict_to_protobuf(meta_dict):
    metadata = metadata_pb2.Metadata()
    metadata.comment = "to be changed"
    metadata.element_count = 1  # will not be touched

    new_metadata = _dict_to_protobuf(dictionary={}, message=metadata)
    results_string = metadata_to_formatted_string(
        message=new_metadata, as_one_line=True
    )
    results_string_expected = metadata_to_formatted_string(
        message=metadata, as_one_line=True
    )
    assert results_string == results_string_expected

    new_metadata = _dict_to_protobuf(dictionary=meta_dict, message=metadata)
    assert isinstance(new_metadata, google.protobuf.message.Message)
    results_string = metadata_to_formatted_string(
        message=new_metadata, as_one_line=True
    )
    results_string_expected = (
        'identifier: "123" comment: "my comment" ancestor_identifiers: "1" '
        'ancestor_identifiers: "2" element_count: 1 relations { identifier: "1" } '
        'experiment { setups { identifier: "1" '
        'optical_units { identifier: "1" detection { '
        'camera { identifier: "1" integration_time { 0.010s } } } } } } '
        'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    )
    assert results_string == results_string_expected

    metadata = metadata_pb2.Metadata()
    metadata.comment = "to be changed"
    metadata.element_count = 1  # will not be touched
    new_metadata = _dict_to_protobuf(
        dictionary=meta_dict, message=metadata, inplace=True
    )
    assert new_metadata is None
    results_string = metadata_to_formatted_string(message=metadata, as_one_line=True)
    results_string_expected = (
        'identifier: "123" comment: "my comment" ancestor_identifiers: "1" '
        'ancestor_identifiers: "2" element_count: 1 relations { identifier: "1" } '
        'experiment { setups { identifier: "1" '
        'optical_units { identifier: "1" detection { '
        'camera { identifier: "1" integration_time { 0.010s } } } } } } '
        'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    )
    assert results_string == results_string_expected


def test_metadata_from_toml_string(metadata_toml):
    result = metadata_from_toml_string(toml_string=None)
    assert result is None

    result = metadata_from_toml_string(metadata_toml)
    assert isinstance(result, dict)
    assert isinstance(result["metadata"], google.protobuf.message.Message)
    results_string = metadata_to_formatted_string(
        message=result["metadata"], as_one_line=True
    )
    results_string_expected = (
        'identifier: "123" comment: "my comment" ancestor_identifiers: "1" '
        'ancestor_identifiers: "2" relations { identifier: "1" } experiment { '
        'setups { identifier: "1" optical_units { identifier: "1" detection { '
        'camera { identifier: "1" integration_time { 0.010s } } } } } } '
        'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    )
    assert results_string == results_string_expected


def test_metadata_from_toml_file(tmp_path, metadata_toml):
    file_path = tmp_path / "metadata_toml.toml"
    with file_path.open("w", encoding="utf-8") as file:
        file.write(metadata_toml)

    result = load_metadata_from_toml(path_or_file_like=None)
    assert result is None

    # load path-like
    result = load_metadata_from_toml(path_or_file_like=file_path)

    assert isinstance(result["metadata"], google.protobuf.message.Message)
    results_string = metadata_to_formatted_string(
        message=result["metadata"], as_one_line=True
    )
    results_string_expected = (
        'identifier: "123" comment: "my comment" ancestor_identifiers: "1" '
        'ancestor_identifiers: "2" relations { identifier: "1" } experiment { '
        'setups { identifier: "1" optical_units { identifier: "1" detection { '
        'camera { identifier: "1" integration_time { 0.010s } } } } } } '
        'localizer { software: "rapidSTORM" } production_time { 2022-05-14T06:58:00Z }'
    )
    assert results_string == results_string_expected

    # load file-like
    with open(file_path, "rb") as f:
        result = load_metadata_from_toml(path_or_file_like=f)
    assert isinstance(result, dict)


def test_message_scheme(metadata):
    scheme = message_scheme(metadata)
    assert isinstance(scheme, dict)
    assert scheme["creation_time"] == "1111-11-11T10:11:11Z"
    assert scheme["modification_time"] == "1970-01-01T00:00:00Z"


def test_merge_metadata(tmp_path, metadata_toml):
    metadata = merge_metadata()
    assert isinstance(metadata, metadata_pb2.Metadata)

    metadata = metadata_pb2.Metadata()
    metadata.comment = "test"
    metadata_2 = dict(comment="other_test")
    metadata_new = merge_metadata(metadata=metadata, other_metadata=metadata_2)
    assert isinstance(metadata_new, metadata_pb2.Metadata)
    assert metadata_new.comment == "other_test"

    metadata = metadata_pb2.Metadata()
    metadata.comment = "test"
    metadata_2 = metadata_pb2.Metadata()
    metadata_2.comment = "other_test"
    metadata_new = merge_metadata(metadata=metadata, other_metadata=metadata_2)
    assert isinstance(metadata_new, metadata_pb2.Metadata)
    assert metadata_new.comment == "other_test"

    file_path = tmp_path / "metadata_toml.toml"
    with file_path.open("w", encoding="utf-8") as file:
        file.write(metadata_toml)
    metadata = metadata_pb2.Metadata()
    metadata.comment = "test"
    metadata_new = merge_metadata(
        metadata=metadata, other_metadata=tmp_path / "metadata_toml.toml"
    )
    assert isinstance(metadata_new, metadata_pb2.Metadata)
    assert metadata_new.comment == "my comment"



"""
Benchmark functions for :mod:`locan.process.register`
"""

import numpy as np

import locan as lc
from locan.process.register import (
    _register_cc_skimage,
    _register_icp_open3d,
    register_icp,
)
from locan.process.transform import transform_affine


class BenchmarkRegister:
    """
    Benchmarks for registering locdata
    """

    def setup(self):
        path = lc.ROOT_DIR / "tests/test_data/five_blobs.txt"
        self.locdata = lc.load_txt_file(path)

        self.offset_true = np.array([100.0, 50.0])
        self.matrix_true = np.array([[1, 0], [0, 1]])
        self.locdata_transformed = transform_affine(
            self.locdata, matrix=self.matrix_true, offset=self.offset_true
        )

    def time__register_cc_skimage(self):
        matrix, offset = _register_cc_skimage(
            locdata=self.locdata, other_locdata=self.locdata_transformed, bin_size=50
        )
        assert np.allclose(np.array(offset), self.offset_true, atol=5)
        assert np.allclose(matrix, self.matrix_true)

    def time__register_icp_open3d(self):
        matrix, offset = _register_icp_open3d(
            points=self.locdata.coordinates,
            other_points=self.locdata_transformed.coordinates,
            verbose=False,
        )
        assert np.allclose(np.array(offset), self.offset_true, atol=5)
        assert np.allclose(matrix, self.matrix_true)

    def time_register_icp(self):
        matrix, offset = register_icp(
            locdata=self.locdata, other_locdata=self.locdata_transformed, verbose=False
        )
        assert np.allclose(np.array(offset), self.offset_true, atol=5)
        assert np.allclose(matrix, self.matrix_true)


def main():
    bm = BenchmarkRegister()
    bm.setup()
    bm.time__register_cc_skimage()
    bm.time__register_icp_open3d()
    bm.time_register_icp()


def main_profile():
    """
    Run benchmarks through profiler.
    """
    import cProfile
    import pstats

    bm = BenchmarkRegister()
    bm.setup()

    profiler = cProfile.Profile()
    profiler.enable()

    for _ in range(100):
        bm.time__register_cc_skimage()
        bm.time__register_icp_open3d()
        bm.time_register_icp()

    profiler.disable()
    stats = pstats.Stats(profiler).sort_stats("cumtime")
    stats.print_stats("time_")


if __name__ == "__main__":
    # main()
    main_profile()



import boost_histogram as bh
import matplotlib.pyplot as plt
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import GroupedPropertyExpectation
from locan.analysis.grouped_property_expectation import (
    GroupedPropertyExpectationResults,
)


class TestGroupedPropertyExpectation:
    def test_init(self, locdata_2d):
        gpe = GroupedPropertyExpectation(meta={"comment": "this is an example"})
        assert (
            str(gpe) == "GroupedPropertyExpectation("
            "loc_property=None, other_loc_property=None, expectation=None)"
        )
        assert gpe.results is None
        assert gpe.meta.comment == "this is an example"

    def test_empty_locdata(self, caplog):
        gpe = GroupedPropertyExpectation().compute(LocData())
        assert gpe.results is None
        gpe.plot()
        gpe.hist()
        # plt.show()

        assert caplog.record_tuples == [
            ("locan.analysis.grouped_property_expectation", 30, "Locdata is empty."),
        ]

        plt.close("all")

    def test_compute(self, locdata_2d):
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        gpe = GroupedPropertyExpectation(
            loc_property="intensity", other_loc_property="localization_count"
        ).compute(locdata=collection)
        # print(gpe.results.values)
        # print(gpe.results.grouped)
        assert isinstance(gpe.results, GroupedPropertyExpectationResults)
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "intensity",
            "localization_count",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "intensity_mean",
            "intensity_std",
            "expectation",
        ]
        assert gpe.results.grouped.expectation.isna().all()

        gpe.plot()
        # plt.show()
        gpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_with_expectation(self, locdata_2d):
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        expected = 2
        gpe = GroupedPropertyExpectation(
            loc_property="intensity",
            other_loc_property="localization_count",
            expectation=expected,
        ).compute(locdata=collection)
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "intensity",
            "localization_count",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "intensity_mean",
            "intensity_std",
            "expectation",
        ]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.intensity / expected
        )

        expected = pd.Series(data=[2], index=[6])
        gpe = GroupedPropertyExpectation(
            loc_property="intensity",
            other_loc_property="localization_count",
            expectation=expected,
        ).compute(locdata=collection)
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "intensity",
            "localization_count",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "intensity_mean",
            "intensity_std",
            "expectation",
        ]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.intensity / 2
        )

        expected = {6: 2}
        gpe = GroupedPropertyExpectation(
            loc_property="intensity",
            other_loc_property="localization_count",
            expectation=expected,
        ).compute(locdata=collection)
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.grouped.index.tolist() == [6]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.intensity / 2
        )

        gpe.plot()
        # plt.show()
        gpe.hist()
        # plt.show()

        plt.close("all")


@pytest.mark.visual
class TestGroupedPropertyExpectationVisual:
    def test_compute(self, locdata_blobs_2d):
        gpe = GroupedPropertyExpectation(
            loc_property="position_x",
            other_loc_property="cluster_label",
            expectation=600,
        ).compute(locdata=locdata_blobs_2d)
        assert isinstance(gpe.results, GroupedPropertyExpectationResults)
        print(gpe.results.values.columns)
        print(gpe.results.grouped)

        gpe.plot()
        plt.show()

        gpe.hist()
        plt.show()

        gpe.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)), log=False)
        plt.show()
        gpe.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)))
        plt.show()

        axes = bh.axis.AxesTuple(
            (
                bh.axis.Regular(20, 3, 503, transform=bh.axis.transform.log),
                bh.axis.Regular(50, 1, 15_000, transform=bh.axis.transform.log),
            )
        )
        gpe.hist(bins=axes)
        plt.show()

        plt.close("all")



"""

Compute Ripley's k function.

Spatial clustering of localization data is characterized by Ripley's k or
related l and h functions [1]_.


Ripley's k function is computed for 2D and 3D data for a series of radii as
described in [2]_ in order to provide evidence for deviations from a spatially
homogeneous Poisson process (i.e. complete spatial randomness, CSR).
Ripley' s k function is estimated by summing all points or over test points
being a random subset of all points:

.. math::

   k(r) = \\frac{1}{\\lambda (n-1)} \\sum_{i=1}^{n} N_{p_{i}}(r)


here :math:`p_i` is the :math:`i^{th}` point of n test points,
:math:`N_{p_{i}}` is the number of points within the region of radius r
around :math:`p_{i}`, and :math:`\\lambda` is the density of all points.

We follow the definition of l and h functions in [2]_. Ripley's l function is:

.. math::

   l(r) &= \\sqrt{k(r)) / \\pi} \\qquad \\text{in 2D}

   l(r) &= \\sqrt[3]{\\frac{3}{4 \\pi} k(r)} \\qquad \\text{in 3D}


And Ripley's h function is:

.. math::

   h(r) = l(r) - r


References
----------
.. [1] B.D. Ripley, Modelling spatial patterns. Journal of the Royal
   Statistical Society, 1977, 172–212.
.. [2] Kiskowski, M. A., Hancock, J. F., and Kenworthy, A. K.,
   On the use of Ripley's K-function and its
   derivatives to analyze domain size.
   Biophysical journal, 2009, 97, 1095–1103.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Literal

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.neighbors import NearestNeighbors

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis

__all__: list[str] = ["RipleysKFunction", "RipleysLFunction", "RipleysHFunction"]

logger = logging.getLogger(__name__)

# The algorithms


def _ripleys_k_function(
    points: npt.ArrayLike,
    radii: Iterable[float],
    region_measure: float = 1,
    other_points: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    """
    Compute Ripley's K function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    points
        2D or 3D points on which to estimate Ripley's K function.
    radii
        Radii at which to compute Ripley's k function.
    region_measure
        region measure (area or volume) for points.
    other_points
        2D or 3D points from which to estimate Ripley's K function (e.g. subset of points). For None other_points
        is set to points (default).

    Returns
    -------
    npt.NDArray[np.float64]
        Ripley's K function evaluated at input radii with shape (n_radii,).
    """
    points = np.asarray(points)
    if other_points is None:
        other_points = points
    else:
        other_points = np.asarray(other_points)

    nn = NearestNeighbors(metric="euclidean").fit(points)

    ripley_0 = []
    for radius in radii:
        indices_list = nn.radius_neighbors(other_points, radius, return_distance=False)
        ripley_0.append(np.array([len(indices) - 1 for indices in indices_list]).sum())

    ripley = (
        region_measure / (len(points) * (len(other_points) - 1)) * np.array(ripley_0)
    )

    return ripley


def _ripleys_l_function(
    points: npt.ArrayLike,
    radii: Iterable[float],
    region_measure: float = 1,
    other_points: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    """
    Evaluates Ripley's L function which is different for 2D and 3D data points.
    For parameter description see _ripleys_k_function.
    """
    if np.shape(points)[1] == 2:
        return_value: npt.NDArray[np.float64] = np.sqrt(
            _ripleys_k_function(points, radii, region_measure, other_points) / np.pi
        )
    elif np.shape(points)[1] == 3:
        return_value = np.cbrt(
            _ripleys_k_function(points, radii, region_measure, other_points)
            * 3
            / 4
            / np.pi
        )
    return return_value


def _ripleys_h_function(
    points: npt.ArrayLike,
    radii: Iterable[float],
    region_measure: float = 1,
    other_points: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    """
    Evaluates Ripley's H function. For parameter description
    see _ripleys_k_function.
    """

    return_value: npt.NDArray[np.float64] = _ripleys_l_function(
        points, radii, region_measure, other_points
    ) - np.asarray(radii)
    return return_value


# The specific analysis classes


class RipleysKFunction(_Analysis):
    """
    Compute Ripley's K function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    radii: Iterable[float] | None
        Radii at which to compute Ripley's k function.
    region_measure : float | Literal['bb']
        Region measure (area or volume) for point region. For 'bb' the region
        measure of the bounding_box is used.


    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with radii as provided and Ripley's K function.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        radii: Iterable[float] | None = None,
        region_measure: float | Literal["bb"] = "bb",
    ) -> None:
        radii = np.linspace(0, 100, 10) if radii is None else radii
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData, other_locdata: LocData | None = None) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data with 2D or 3D coordinates on which to estimate
            Ripley's K function.
        other_locdata
            Other localization data from which to estimate Ripley's K function
            (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter["region_measure"] == "bb":
            region_measure = float(locdata.properties["region_measure_bb"])
        else:
            region_measure = self.parameter["region_measure"]

        ripley = _ripleys_k_function(
            points=points,
            radii=self.parameter["radii"],
            region_measure=region_measure,
            other_points=other_points,
        )
        self.results = pd.DataFrame(
            {"radius": self.parameter["radii"], "Ripley_k_data": ripley}
        )
        self.results = self.results.set_index("radius")
        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        return plot(self=self, ax=ax, **kwargs)


class RipleysLFunction(_Analysis):
    """
    Compute Ripley's L function for two- or three-dimensional data at the
    given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    radii: npt.ArrayLike | None
        Radii at which to compute Ripley's k function.
    region_measure : float |  Literal["bb"]
        Region measure (area or volume) for point region. For 'bb' the region
        measure of the bounding_box is used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        Data frame with radii as provided and Ripley's L function.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        radii: npt.ArrayLike | None = None,
        region_measure: float | Literal["bb"] = "bb",
    ) -> None:
        radii = np.linspace(0, 100, 10) if radii is None else radii
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData, other_locdata: LocData | None = None) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data with 2D or 3D coordinates on which to estimate
            Ripley's L function.
        other_locdata
            Other localization data from which to estimate Ripley's L function
            (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter["region_measure"] == "bb":
            region_measure = float(locdata.properties["region_measure_bb"])
        else:
            region_measure = self.parameter["region_measure"]

        ripley = _ripleys_l_function(
            points=points,
            radii=self.parameter["radii"],
            region_measure=region_measure,
            other_points=other_points,
        )
        self.results = pd.DataFrame(
            {"radius": self.parameter["radii"], "Ripley_l_data": ripley}
        )
        self.results = self.results.set_index("radius")
        return self

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        return plot(self, ax, **kwargs)


class RipleysHFunction(_Analysis):
    """
    Compute Ripley's H function for two- or three-dimensional data at the given radii.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    radii: npt.ArrayLike | None
        Radii at which to compute Ripley's k function.
    region_measure : float | Literal['bb']
        Region measure (area or volume) for point region. For 'bb' the region
        measure of the bounding_box is used.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict[str, Any]
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pd.DataFrame | None
        Data frame with radii as provided and Ripley's H function.
    Ripley_h_maximum : pd.DataFrame | None
        Data frame with radius and Ripley's H value for the radius at which
        the H function has its maximum.
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        radii: npt.ArrayLike | None = None,
        region_measure: float | Literal["bb"] = "bb",
    ) -> None:
        radii = np.linspace(0, 100, 10) if radii is None else radii
        parameters: dict[str, Any] = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results: pd.DataFrame | None = None
        self._Ripley_h_maximum: pd.DataFrame | None = None

    def compute(self, locdata: LocData, other_locdata: LocData | None = None) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data with 2D or 3D coordinates on which to estimate
            Ripley's H function.
        other_locdata
            Other localization data from which to estimate Ripley's H function
            (e.g. subset of points).
            For None other_points is set to points (default).

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        # reset secondary results
        self._Ripley_h_maximum = None

        points = locdata.coordinates
        if other_locdata is not None:
            other_points = other_locdata.coordinates
        else:
            other_points = None

        # choose the right region_measure
        # todo: add other hull regions
        if self.parameter["region_measure"] == "bb":
            region_measure = float(locdata.properties["region_measure_bb"])
        else:
            region_measure = self.parameter["region_measure"]

        ripley = _ripleys_h_function(
            points=points,
            radii=self.parameter["radii"],
            region_measure=region_measure,
            other_points=other_points,
        )
        self.results = pd.DataFrame(
            {"radius": self.parameter["radii"], "Ripley_h_data": ripley}
        )
        self.results = self.results.set_index("radius")
        return self

    @property
    def Ripley_h_maximum(self) -> pd.DataFrame | None:
        if self.results is None:
            self._Ripley_h_maximum = None
        elif self._Ripley_h_maximum is None:
            index = self.results["Ripley_h_data"].idxmax()
            self._Ripley_h_maximum = pd.DataFrame(
                {"radius": index, "Ripley_h_maximum": self.results.loc[index]}
            )
        return self._Ripley_h_maximum

    @Ripley_h_maximum.deleter
    def Ripley_h_maximum(self) -> None:
        self._Ripley_h_maximum = None

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        return plot(self, ax, **kwargs)


# Interface functions


def plot(
    self: _Analysis, ax: mpl.axes.Axes | None = None, **kwargs: Any
) -> mpl.axes.Axes:
    """
    Provide plot of results as :class:`matplotlib.axes.Axes` object.

    Parameters
    ----------
    ax
        The axes on which to show the image
    kwargs
        Other parameters passed to :func:`matplotlib.pyplot.plot`.

    Returns
    -------
    matplotlib.axes.Axes
        Axes object with the plot.
    """
    if ax is None:
        ax = plt.gca()

    if self.results is None:
        return ax

    self.results.plot(ax=ax, **kwargs)

    if self.results.columns[0] == "Ripley_k_data":
        ylabel = "K-function"
    elif self.results.columns[0] == "Ripley_l_data":
        ylabel = "L-function"
    elif self.results.columns[0] == "Ripley_h_data":
        ylabel = "H-function"
    else:
        ylabel = ""

    ax.set(title="Ripley's " + ylabel, xlabel="radius", ylabel=ylabel)

    return ax



"""
Utility methods to print system and dependency information.

adapted from :func:`pandas.show_versions` (`locan/licences/PANDAS.rst`)
(BSD 3-Clause License, Copyright (c) 2008-2011, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development
Team, Copyright (c) 2011-2021, Open source contributors.)
and from :func:`scikit-learn.show_versions` (`locan/licences/SCIKIT-LEARN.rst`)
(BSD 3-Clause License, Copyright (c) 2007-2020 The scikit-learn developers.)
"""

from __future__ import annotations

import importlib
import locale
import os
import platform
import struct
import sys
from typing import Any

from locan import __version__ as locan_version
from locan.dependencies import EXTRAS_REQUIRE, INSTALL_REQUIRES

__all__: list[str] = ["system_info", "dependency_info", "show_versions"]


def system_info(verbose: bool = True) -> dict[str, Any]:
    """
    Return system information.

    Parameters
    ----------
    verbose
        If True information on node and executable path are added.

    Return
    ------
    dict
        System information
    """
    uname_result = platform.uname()
    language_code, encoding = locale.getlocale()

    sys_info = {
        "python-bits": struct.calcsize("P") * 8,
        "system": uname_result.system,
        "release": uname_result.release,
        "version": uname_result.version,
        "machine": uname_result.machine,
        "processor": uname_result.processor,
        "byteorder": sys.byteorder,
        "LC_ALL": os.environ.get("LC_ALL"),
        "LANG": os.environ.get("LANG"),
        "LOCALE": {"language-code": language_code, "encoding": encoding},
    }

    if verbose:
        sys_info.update({"node": uname_result.node, "executable": sys.executable})

    return sys_info


def dependency_info(
    extra_dependencies: bool = True, other_dependencies: list[str] | None = None
) -> dict[str, Any]:
    """
    Overview of the installed version of main dependencies.

    Parameters
    ----------
    extra_dependencies
        Include extra dependencies

    other_dependencies
        Include other module names.

    Returns
    -------
    dict
        Version information on relevant Python libraries
    """
    deps = INSTALL_REQUIRES.copy()

    if extra_dependencies:
        deps = deps.union(EXTRAS_REQUIRE)

    if other_dependencies:
        deps = deps.union(other_dependencies)

    deps_info: dict[str, str | None] = {}
    for modname in deps:
        try:
            deps_info[modname] = importlib.metadata.version(modname)
        except importlib.metadata.PackageNotFoundError:
            deps_info[modname] = None
    return dict(sorted(deps_info.items()))


def show_versions(
    locan: bool = True,
    python: bool = True,
    system: bool = True,
    dependencies: bool = True,
    verbose: bool = True,
    extra_dependencies: bool = True,
    other_dependencies: list[str] | None = None,
) -> None:
    """
    Print useful debugging information on system and dependency versions.

    Parameters
    ----------
    locan
        Show locan version
    python
        Show python version
    system
        Show system information
    verbose
        If True information on node and executable path are added.
    dependencies
        Show main dependencies
    extra_dependencies
        Include extra dependencies as specified in setup.py if True.
    other_dependencies
        Include other module names.
    """

    locan_info = {"version": locan_version}
    python_info = {"version": platform.python_version()}
    sys_info = system_info(verbose)
    deps_info = dependency_info(extra_dependencies, other_dependencies)

    if locan:
        print("\nLocan:")
        for key, value in locan_info.items():
            print(f"{key:>10}: {value}")

    if python:
        print("\nPython:")
        for key, value in python_info.items():
            print(f"{key:>10}: {value}")

    if system:
        print("\nSystem:")
        for key, value in sys_info.items():
            print(f"{key:>10}: {value}")

    if dependencies:
        print("\nPython dependencies:")
        for key, value in deps_info.items():
            print(f"{key:>10}: {value}")



"""

Transform localization data with a BUnwarpJ transformation matrix.

This module provides functions to transform coordinates in LocData objects by
applying a B-spline transformation as defined with the ImageJ/Fiji
plugin BunwarpJ_ [1]_, [2]_.

.. _BunwarpJ: https://imagej.net/BUnwarpJ

References
----------

.. [1] I. Arganda-Carreras, C. O. S. Sorzano, R. Marabini, J.-M. Carazo,
   C. Ortiz-de Solorzano, and J. Kybic,
   "Consistent and Elastic Registration of Histological Sections using
   Vector-Spline Regularization",
   Lecture Notes in Computer Science, Springer Berlin / Heidelberg,
   volume 4241/2006,
   CVAMIA: Computer Vision Approaches to Medical Image Analysis,
   pages 85-95, 2006.

.. [2] C.Ó. Sánchez Sorzano, P. Thévenaz, M. Unser,
   "Elastic Registration of Biological Images Using Vector-Spline
   Regularization",
   IEEE Transactions on Biomedical Engineering, vol. 52, no. 4,
   pages 652-663, 2005.

"""

from __future__ import annotations

import os
import sys
from itertools import islice
from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd
from scipy import interpolate

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.process.transform.spatial_transformation import transform_affine

__all__: list[str] = ["bunwarp"]


def _unwarp(
    points: npt.ArrayLike,
    matrix_x: npt.ArrayLike,
    matrix_y: npt.ArrayLike,
    pixel_size: tuple[float, float],
) -> npt.NDArray[np.float64]:
    """
    Transform points with raw matrix from BunwarpJ.

    Parameters
    ----------
    points
        Point coordinates to be transformed.
        Array with shape (n_points, 2).
    matrix_x
        Transformation matrix for x coordinates
    matrix_y
        Transformation matrix for y coordinates
    pixel_size
        Pixel size for x and y component as used in ImageJ for registration

    Returns
    -------
    npt.NDArray[np.float64]
        Transformed point coordinates with shape (n_points, 2).
    """
    points_ = np.asarray(points)
    point_indices = np.divide(points_, pixel_size)

    matrix_x = np.asarray(matrix_x)
    matrix_y = np.asarray(matrix_y)
    if matrix_x.shape == matrix_y.shape:
        matrix_size = matrix_x.shape
    else:
        raise TypeError("matrix_x and matrix_y must have the same shape.")

    x = np.arange(matrix_size[0])
    y = np.arange(matrix_size[1])
    rgi_x = interpolate.RegularGridInterpolator(points=(x, y), values=matrix_x)
    rgi_y = interpolate.RegularGridInterpolator(points=(x, y), values=matrix_y)
    matrix_x_interpolated = rgi_x(point_indices, method="linear")
    matrix_y_interpolated = rgi_y(point_indices, method="linear")

    new_point_indices = np.stack([matrix_x_interpolated, matrix_y_interpolated], axis=1)
    new_points: npt.NDArray[np.float64] = np.multiply(new_point_indices, pixel_size)
    return new_points


def _read_matrix(
    path: str | os.PathLike[Any],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Read file with raw matrix from BunwarpJ.

    Parameters
    ----------
    path
        Path to file with a raw matrix from BunwarpJ.

    Returns
    -------
    tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]
        x transformation array, y transformation array
    """
    with open(path) as file:
        header = list(islice(file, 2))

    # Get image heigth and width
    width = int(header[0].split("=")[1])
    height = int(header[1].split("=")[1])
    matrix_size = np.array([width, height])

    matrix_x = pd.read_csv(
        path, skiprows=4, header=None, nrows=matrix_size[1], sep=r"\s+"
    ).values.T  # transform values to get array[x, y]
    matrix_y = pd.read_csv(
        path, skiprows=(6 + matrix_size[1]), header=None, sep=r"\s+"
    ).values.T  # transform values to get array[x, y]

    return matrix_x, matrix_y


def bunwarp(
    locdata: LocData,
    matrix_path: str | os.PathLike[Any],
    pixel_size: tuple[float, float],
    flip: bool = False,
) -> LocData:
    """
    Transform coordinates by applying a B-spline transformation
    as represented by a raw transformation matrix from BunwarpJ.

    Parameters
    ----------
    locdata
        specifying the localization data on which to perform the manipulation.
    matrix_path
        Path to file with a raw matrix from BunwarpJ.
    pixel_size
        Pixel sizes used to determine transition matrix in ImageJ
    flip
        Flip locdata along x-axis before transformation

    Returns
    -------
    LocData
        New localization data with transformed coordinates.
    """
    local_parameter = locals()

    matrix_x, matrix_y = _read_matrix(matrix_path)

    if flip:
        image_size = np.multiply(matrix_x.shape, pixel_size)
        locdata = transform_affine(  # type: ignore[assignment]
            locdata, matrix=[[-1, 0], [0, 1]], offset=[image_size[0], 0]
        )

    new_points = _unwarp(locdata.coordinates, matrix_x, matrix_y, pixel_size)

    # new LocData object
    new_dataframe = locdata.data.copy()
    df = pd.DataFrame(
        {"position_x": new_points[:, 0], "position_y": new_points[:, 1]},
        index=locdata.data.index,
    )
    # cast dtypes
    columns_intersection = list(set(df.columns) & set(new_dataframe.columns))
    df = df.astype(new_dataframe[columns_intersection].dtypes)

    new_dataframe.update(df)
    new_locdata = LocData.from_dataframe(new_dataframe)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata



"""
This LocData test is placed in a dedicated file to get rid of the following
problem:

Under GitHub actions with windows and python 3.10 the following error occurred
sporadically:
>       assert count_0 + 2 == count_2
E       assert (19 + 2) == 15
locan\\tests\\data\\test_locdata.py:60: AssertionError
"""

import copy

from locan import LocData


def test_LocData_count():
    count_0 = copy.deepcopy(LocData.count)
    dat = LocData()
    count_1 = copy.deepcopy(LocData.count)
    dat_2 = LocData()
    count_2 = copy.deepcopy(LocData.count)
    assert dat.properties == dat_2.properties
    del dat
    count_3 = copy.deepcopy(LocData.count)
    assert count_0 + 1 == count_1
    assert count_0 + 2 == count_2
    assert count_2 - 1 == count_3



"""

Utility functions for scripts.

"""

from __future__ import annotations

import re


def _type_converter_rescale(input_string: str) -> str | bool | tuple[float] | None:
    if input_string == "None":
        return None
    elif input_string == "True":
        return True
    elif input_string == "False":
        return False
    else:
        pattern = re.match(
            r"\(?([0-9]*[.]?[0-9]+),?\s?([0-9]*[.]?[0-9]+)\)?", input_string
        )
        if pattern:
            return tuple(float(element) for element in pattern.groups())  # type: ignore[return-value]
        else:
            return input_string



import logging

import boost_histogram as bh
import numpy as np
import pytest

from locan import Bins, histogram
from locan.process.aggregate import (
    _accumulate_1d,
    _accumulate_2d,
    _bin_edges_to_bin_centers,
    _bin_edges_to_bin_size,
    _bin_edges_to_bin_size_one_dimension,
    _bin_edges_to_n_bins,
    _bin_edges_to_n_bins_one_dimension,
    _bin_size_to_bin_edges_one_dimension,
    _BinsFromBoostHistogramAxis,
    _BinsFromEdges,
    _BinsFromNumber,
    _BinsFromSize,
    _indices_to_bin_centers,
    _is_1d_array_of_scalar,
    _is_1d_array_of_two_scalar,
    _is_2d_homogeneous_array,
    _is_2d_inhomogeneous_array,
    _is_2d_inhomogeneous_array_of_1d_array_of_scalar,
    _is_scalar,
    _is_single_element,
    _n_bins_to_bin_edges_one_dimension,
)

data_scalars = {"1": 1, "()": ()}

data_tuples = {
    "((),)": ((),),
    "(1,)": (1,),
    "(1, 2)": (1, 2),
    "(1, 2, 3)": (1, 2, 3),
    "((1, 2),)": ((1, 2),),
    "((1, 2), (1, 2))": ((1, 2), (1, 2)),
    "((1, 2), (1, 2, 3))": ((1, 2), (1, 2, 3)),
    "((1, 2), ((1, 2), (1, 2)))": ((1, 2), ((1, 2), (1, 2))),
    "(1, (1, 2))": (1, (1, 2)),
}

data_lists = {
    "[[]]": [[]],
    "[1]": [1],
    "[1, 2]": [1, 2],
    "[1, 2, 3]": [1, 2, 3],
    "[[1, 2]]": [[1, 2]],
    "[[1, 2], [1, 2]]": [[1, 2], [1, 2]],
    "[[1, 2], [1, 2, 3]]": [[1, 2], [1, 2, 3]],
    "[[1, 2], [[1, 2], [1, 2]]]": [[1, 2], [[1, 2], [1, 2]]],
    "[1, [1, 2]]": [1, [1, 2]],
}

data_ndarrays = {
    "np.array((1))": np.array(1),
    "np.array((1,))": np.array((1,)),
    "np.array((1, 2))": np.array((1, 2)),
    "np.array([(1, 2)])": np.array([(1, 2)]),
    "np.array([(1, 2)], dtype=object)": np.array([(1, 2)], dtype=object),
    "np.array([(1, 2), (1, 2, 3)], dtype=object)": np.array(
        [(1, 2), (1, 2, 3)], dtype=object
    ),
}

data_all = {**data_scalars, **data_tuples, **data_lists, **data_ndarrays}

expect_is_scalar = ["1", "np.array((1))"]

expect_is_single_element = ["1", "(1,)", "[1]", "np.array((1))", "np.array((1,))"]

expect_is_1d_array_of_scalar = [
    "(1, 2)",
    "(1, 2, 3)",
    "[1, 2]",
    "[1, 2, 3]",
    "np.array((1, 2))",
    "(1,)",
    "[1]",
    "np.array((1,))",
]

expect_is_1d_array_of_two_scalar = ["(1, 2)", "[1, 2]", "np.array((1, 2))"]

expect_is_2d_homogeneous_array = [
    "((1, 2),)",
    "((1, 2), (1, 2))",
    "[[1, 2]]",
    "[[1, 2], [1, 2]]",
    "np.array([(1, 2)])",
]

expect_is_2d_inhomogeneous_array = [
    "((1, 2), (1, 2, 3))",
    "(1, (1, 2))",
    "[[1, 2], [1, 2, 3]]",
    "[1, [1, 2]]",
    "np.array([(1, 2), (1, 2, 3)], dtype=object)",
]

expect_is_2d_inhomogeneous_array_of_1d_array_of_scalar = [
    "((1, 2), (1, 2, 3))",
    "[[1, 2], [1, 2, 3]]",
    "np.array([(1, 2), (1, 2, 3)], dtype=object)",
]


def test__is_scalar():
    for key, value in data_all.items():
        if key in expect_is_scalar:
            assert _is_scalar(value)
        else:
            assert not _is_scalar(value)


def test__is_single_element():
    for key, value in data_all.items():
        if key in expect_is_single_element:
            assert _is_single_element(value)
        else:
            assert not _is_single_element(value)


def test__is_1d_array_of_scalar():
    for key, value in data_all.items():
        if key in expect_is_1d_array_of_scalar:
            assert _is_1d_array_of_scalar(value)
        else:
            assert not _is_1d_array_of_scalar(value)


def test__is_1d_array_of_two_scalar():
    for key, value in data_all.items():
        if key in expect_is_1d_array_of_two_scalar:
            assert _is_1d_array_of_two_scalar(value)
        else:
            assert not _is_1d_array_of_two_scalar(value)


def test__is_2d_homogeneous_array():
    for key, value in data_all.items():
        if key in expect_is_2d_homogeneous_array:
            assert _is_2d_homogeneous_array(value)
        else:
            assert not _is_2d_homogeneous_array(value)


def test__is_2d_inhomogeneous_array():
    for key, value in data_all.items():
        if key in expect_is_2d_inhomogeneous_array:
            assert _is_2d_inhomogeneous_array(value)
        else:
            assert not _is_2d_inhomogeneous_array(value)


def test__is_2d_inhomogeneous_array_of_1d_array_of_scalar():
    for key, value in data_all.items():
        if key in expect_is_2d_inhomogeneous_array_of_1d_array_of_scalar:
            assert _is_2d_inhomogeneous_array_of_1d_array_of_scalar(value)
        else:
            assert not _is_2d_inhomogeneous_array_of_1d_array_of_scalar(value)


def test__n_bins_to_bin_edges_one_dimension():
    bin_edges = _n_bins_to_bin_edges_one_dimension(10, (10, 20))
    assert bin_edges.shape == (11,)


def test__bin_size_to_bin_edges_one_dimension():
    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=None)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9))

    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (-1, 10), extend_range=None)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (-1, 3, 7))

    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=True)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9, 13))
    bin_edges = _bin_size_to_bin_edges_one_dimension(4, (1, 10), extend_range=False)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 5, 9, 10))

    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=None)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 10))
    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=True)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 21))
    bin_edges = _bin_size_to_bin_edges_one_dimension(20, (1, 10), extend_range=False)
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 10))

    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (1, 2, 3, 3, 2), (1, 11), extend_range=None
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10))
    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (1, 2, 3, 3, 2), (1, 11), extend_range=True
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10, 12))
    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (1, 2, 3, 3, 2), (1, 11), extend_range=False
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2, 4, 7, 10, 11))

    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (10, 20, 30), (1, 2), extend_range=None
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2))
    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (10, 20, 30), (1, 2), extend_range=True
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 11))
    bin_edges = _bin_size_to_bin_edges_one_dimension(
        (10, 20, 30), (1, 2), extend_range=False
    )
    assert isinstance(bin_edges, np.ndarray)
    assert np.array_equal(bin_edges, (1, 2))

    with pytest.raises(TypeError):
        _bin_size_to_bin_edges_one_dimension(((4,), 2), (0, 10), extend_range=False)


def test__bin_edges_to_n_bins_one_dimension():
    n_bins = _bin_edges_to_n_bins_one_dimension((1, 3, 5))
    assert n_bins == 2
    n_bins = _bin_edges_to_n_bins_one_dimension([1, 2, 4])
    assert n_bins == 2


def test__bin_edges_to_n_bins():
    n_bins = _bin_edges_to_n_bins([1, 3, 5])
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins(([1, 3, 5],))
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins([1, 2, 4])
    assert n_bins == (2,)
    n_bins = _bin_edges_to_n_bins(((1, 3, 5), (1, 2, 4, 5)))
    assert np.array_equal(n_bins, (2, 3))
    n_bins = _bin_edges_to_n_bins([[1, 3, 5], [1, 2, 3, 4]])
    assert np.array_equal(n_bins, (2, 3))
    n_bins = _bin_edges_to_n_bins(np.array([[1, 3, 5], [1, 2, 3]]))
    assert np.array_equal(n_bins, (2, 2))


def test__bin_edges_to_bin_size_one_dimension():
    bin_size = _bin_edges_to_bin_size_one_dimension((1, 3, 5))
    assert bin_size == 2
    bin_size = _bin_edges_to_bin_size_one_dimension([1, 2, 4])
    assert np.array_equal(bin_size, (1, 2))
    bin_size = _bin_edges_to_bin_size_one_dimension([1, 2])
    assert bin_size == 1


def test__bin_edges_to_bin_size(caplog):
    bin_size = _bin_edges_to_bin_size([1, 3, 5])
    assert bin_size == (2,)
    bin_size = _bin_edges_to_bin_size([-1, 1, 3.001, 5])
    assert np.array_equal(bin_size[0], [2, 2.001, 1.999])
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)
    bin_size = _bin_edges_to_bin_size([-1, 1, 3.00001, 5])
    assert bin_size == (2,)
    assert np.array_equal(
        caplog.record_tuples,
        [
            (
                "locan.process.aggregate",
                10,
                "bin_sizes differ by floating point instability with less than rtol=1.e-5",
            )
        ],
    )
    bin_size = _bin_edges_to_bin_size(([1, 3, 5],))
    assert bin_size == (2,)
    bin_size = _bin_edges_to_bin_size([1, 2, 4])
    assert np.array_equal(bin_size[0], (1, 2))
    bin_size = _bin_edges_to_bin_size(((1, 3, 5), (1, 2, 4, 5)))
    assert bin_size[0] == 2
    assert np.array_equal(bin_size[1], (1, 2, 1))
    bin_size = _bin_edges_to_bin_size([[1, 3, 5], [1, 2, 3, 4]])
    assert bin_size == (2, 1)
    bin_size = _bin_edges_to_bin_size(np.array([[1, 3, 5], [1, 2, 3]]))
    assert bin_size == (2, 1)


def test__bin_edges_to_bin_centers():
    bin_centers = _bin_edges_to_bin_centers([1, 3, 5])
    assert np.array_equal(bin_centers, ((2, 4),))

    bin_centers = _bin_edges_to_bin_centers(([1, 3, 5],))
    assert np.array_equal(bin_centers, ((2, 4),))

    bin_centers = _bin_edges_to_bin_centers(((1, 3, 5), (1, 2, 4, 6)))
    expected = ((2, 4), (1.5, 3, 5))
    for bc, ex in zip(bin_centers, expected):
        assert np.array_equal(bc, ex)

    bin_edges = np.array([[0, 1, 2, 4, 8, 9], [0, 1, 4, 8]], dtype=object)
    bin_centers = _bin_edges_to_bin_centers(bin_edges)
    expected = (np.array([0.5, 1.5, 3.0, 6, 8.5]), np.array([0.5, 2.5, 6]))
    for bc, ex in zip(bin_centers, expected):
        assert np.array_equal(bc, ex)


def test__indices_to_bin_centers():
    indices = 2
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = 3
    assert np.array_equal(bin_centers, expected)

    indices = np.array([0, 2, 1])
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([0.5, 3, 1.5])
    assert np.array_equal(bin_centers, expected)

    indices = np.array([[0, 1], [2, 2], [4, 3]])
    bin_edges = np.array([0, 1, 2, 4, 8, 9])
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([[0.5, 1.5], [3, 3], [8.5, 6]])
    assert np.array_equal(bin_centers, expected)

    indices = np.array([[0, 1], [2, 2], [4, 3]])
    bin_edges = np.array([[0, 1, 2, 4, 8, 9], [1, 2, 4, 8, 9]], dtype=object)
    bin_centers = _indices_to_bin_centers(bin_edges, indices)
    expected = np.array([[0.5, 3], [3, 6], [8.5, 8.5]])
    assert np.array_equal(bin_centers, expected)


def test__BinsFromBoostHistogramAxis():
    bhaxis = bh.axis.Regular(5, 0, 10)
    bins = _BinsFromBoostHistogramAxis(bins=bhaxis)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert np.array_equal(bins.bin_size[0], (2, 2, 2, 2, 2))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))

    bhaxes = bh.axis.AxesTuple((bh.axis.Regular(5, 0, 10), bh.axis.Regular(2, 0, 10)))
    bins = _BinsFromBoostHistogramAxis(bins=bhaxes)
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 2)
    assert np.array_equal(bins.bin_size[0], (2.0, 2.0, 2.0, 2.0, 2.0))
    assert np.array_equal(bins.bin_size[1], (5.0, 5.0))
    expected_edges = [np.array([0, 2, 4, 6, 8, 10]), np.array([0, 5, 10])]
    for bin_edges, edges in zip(bins.bin_edges, expected_edges):
        assert np.array_equal(bin_edges, edges)
    expected_centers = [np.array([1, 3, 5, 7, 9]), np.array([2.5, 7.5])]
    for bin_centers, expected_cents in zip(bins.bin_centers, expected_centers):
        assert np.array_equal(bin_centers, expected_cents)


def test__BinsFromEdges():
    bins = _BinsFromEdges(bin_edges=(0, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 4.0),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))

    bins = _BinsFromEdges(bin_edges=((0, 2, 4),))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 4.0),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))

    bins = _BinsFromEdges(bin_edges=(1, 2, 5))
    assert bins.dimension == 1
    assert bins.bin_range == ((1, 5),)
    assert bins.n_bins == (2,)
    assert np.array_equal(bins.bin_size[0], np.array([1, 3]))
    assert np.array_equal(bins.bin_edges[0], np.array([1, 2, 5]))

    bins = _BinsFromEdges(bin_edges=((0, 2, 4), (1, 2, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 4.0), (1.0, 5.0))
    assert bins.n_bins == (2, 2)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], np.array([1, 3]))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))
    assert np.array_equal(bins.bin_edges[-1], np.array([1, 2, 5]))

    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=())
    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=[(0, 2, 4), ((0, 2, 4), (0, 2, 4))])
    with pytest.raises(TypeError):
        _BinsFromEdges(bin_edges=[(0, 2, 4), 2])


def test__BinsFromNumber_():
    bins = _BinsFromNumber(n_bins=5, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromNumber(n_bins=(5,), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromNumber(n_bins=(2, 5), bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 5)
    assert bins.bin_size == (5, 1)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 2, 3, 4, 5]))

    bins = _BinsFromNumber(n_bins=2, bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 2)
    assert bins.bin_size == (5, 2.5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2.5, 5]))

    bins = _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(0, 10))
    assert bins.dimension == 3
    assert bins.bin_range == ((0, 10), (0, 10), (0, 10))
    assert bins.n_bins == (2, 5, 2)
    assert bins.bin_size == (5, 2, 5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[2], np.array([0, 5, 10]))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=5, bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5, (1, 2)), bin_range=(1, 2, 3))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5,), bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2,), bin_range=((0, 10), (0, 5)))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=1)
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(2, 5, 2), bin_range=(1, 2, 3))

    with pytest.raises(TypeError):
        _BinsFromNumber(n_bins=(5, (1, 2)), bin_range=(1, 2))


def test__BinsFromSize():
    bins = _BinsFromSize(bin_size=2, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromSize(bin_size=(2,), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))

    bins = _BinsFromSize(bin_size=((1, 2, 3, 4),), bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (4,)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=3, bin_range=(0, 10), extend_range=False)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (4,)
    assert np.array_equal(bins.bin_size[0], (3.0, 3.0, 3.0, 1.0))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 3, 6, 9, 10]))

    bins = _BinsFromSize(bin_size=(5, 1), bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 5))
    assert bins.n_bins == (2, 5)
    assert bins.bin_size == (5, 1)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 5, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 2, 3, 4, 5]))

    bins = _BinsFromSize(bin_size=2, bin_range=((0, 10), (0, 5)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 4))
    assert bins.n_bins == (5, 2)
    assert bins.bin_size == (2, 2)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 2, 4]))

    bins = _BinsFromSize(bin_size=(2, 5), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0, 10), (0, 10))
    assert bins.n_bins == (5, 2)
    assert bins.bin_size == (2, 5)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 5, 10]))

    bins = _BinsFromSize(bin_size=((1, 2, 3, 4), (1, 2, 3, 1)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 7.0))
    assert bins.n_bins == (4, 4)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 1))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 7]))

    bins = _BinsFromSize(
        bin_size=((1, 2, 3, 4), (1, 2, 3, 4)), bin_range=((0, 10), (0, 10))
    )
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (4, 4)
    assert np.array_equal(bins.bin_size[0], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 3, 6, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=(2, (1, 2, 3, 4)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 4)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    bins = _BinsFromSize(bin_size=(2, (1, 2, 3, 4)), bin_range=((0, 10), (0, 20)))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 10.0))
    assert bins.n_bins == (5, 4)
    assert bins.bin_size[0] == 2
    assert np.array_equal(bins.bin_size[1], (1, 2, 3, 4))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_edges[1], np.array([0, 1, 3, 6, 10]))

    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=5, bin_range=(0,))
    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=5, bin_range=(1, 2, 3))
    with pytest.raises(TypeError):
        _BinsFromSize(bin_size=(2, 2, (1, 2, 3, 4)), bin_range=((0, 10), (0, 20)))
    with pytest.raises(TypeError):
        _BinsFromSize(
            bin_size=((1, 2, 3, 4), (1, 2, 3, 4), (1, 2, 3, 4)),
            bin_range=((0, 10), (0, 20)),
        )


def test_Bins():
    bins = Bins(bin_edges=(0, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (2,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(2, 0, 4),)

    bins = Bins(n_bins=5, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(5, 0, 10),)

    bins = Bins(bin_size=2, bin_range=(0, 10))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(5, 0, 10),)

    bins = Bins(bin_size=(2, (1, 2, 3)), bin_range=(0, 10))
    assert bins.dimension == 2
    assert bins.bin_range == ((0.0, 10.0), (0.0, 6.0))
    assert bins.n_bins == (5, 3)
    assert np.array_equal(bins.bin_size[0], 2)
    assert np.array_equal(bins.bin_size[1], (1, 2, 3))
    assert bins.is_equally_sized == (True, False)
    assert bins.boost_histogram_axes == (
        bh.axis.Regular(5, 0, 10),
        bh.axis.Variable([0, 1, 3, 6]),
    )

    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10)))
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(5, 0, 10),)

    bins = Bins(n_bins=(2, 5), bin_range=(0, 10), labels=["position_x", "position_y"])
    assert bins.labels == ["position_x", "position_y"]
    assert bins.dimension == 2
    assert bins.is_equally_sized == (True, True)
    assert bins.boost_histogram_axes == (
        bh.axis.Regular(2, 0, 10),
        bh.axis.Regular(5, 0, 10),
    )
    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10)), labels=["position_x"])
    assert bins.labels == ["position_x"]
    assert bins.dimension == 1
    bins = Bins(bins=Bins(n_bins=5, bin_range=(0, 10), labels=["position_x"]))
    assert bins.labels == ["position_x"]
    assert bins.dimension == 1
    with pytest.raises(ValueError):
        Bins(n_bins=5, bin_range=(0, 10), labels=["position_x", "position_y"])


def test_Bins_with_boost_histogram():
    bhaxis = bh.axis.Regular(5, 0, 10)
    bins = Bins(bins=bhaxis)
    assert bins.dimension == 1
    assert bins.bin_range == ((0.0, 10.0),)
    assert bins.n_bins == (5,)
    assert np.array_equal(bins.bin_size[0], (2, 2, 2, 2, 2))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 2, 4, 6, 8, 10]))
    assert np.array_equal(bins.bin_centers[0], np.array([1, 3, 5, 7, 9]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(5, 0, 10),)


def test_Bins_methods():
    bins = Bins(bin_edges=(0, 1, 2, 4))
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (3,)
    assert np.array_equal(bins.bin_size[0], (1, 1, 2))
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 2, 4]))
    assert bins.is_equally_sized == (False,)
    assert bins.boost_histogram_axes == (bh.axis.Variable([0, 1, 2, 4]),)

    bins = Bins(bin_edges=(0, 1, 2, 4)).equalize_bin_size()
    assert bins.dimension == 1
    assert bins.bin_range == ((0, 4),)
    assert bins.n_bins == (4,)
    assert bins.bin_size == (1,)
    assert np.array_equal(bins.bin_edges[0], np.array([0, 1, 2, 3, 4]))
    assert bins.is_equally_sized == (True,)
    assert bins.boost_histogram_axes == (bh.axis.Regular(4, 0, 4),)

    bins = Bins(
        bin_edges=((-1.0, 1.0, 2.0, 4.0), (-1.0, 1.0, 2.0, 6.0))
    ).equalize_bin_size()
    assert bins.dimension == 2
    assert bins.bin_range == ((-1, 3), (-1, 5))
    assert bins.n_bins == (2, 3)
    assert bins.bin_size == (2, 2)
    assert np.array_equal(bins.bin_edges[0], np.array([-1, 1, 3]))
    assert bins.is_equally_sized == (True, True)
    assert bins.boost_histogram_axes == (
        bh.axis.Regular(2, -1, 3),
        bh.axis.Regular(3, -1, 5),
    )

    bins = Bins(bin_size=2.0, bin_range=(-1, 10)).equalize_bin_size()
    assert bins.dimension == 1
    assert bins.bin_range == ((-1.0, 9),)
    assert bins.n_bins == (5,)
    assert bins.bin_size == (2,)


def test_histogram(locdata_blobs_2d):
    hist = histogram(locdata_blobs_2d, n_bins=10)
    assert hist.labels == ["position_x", "position_y", "counts"]
    assert hist.data.dtype == "float64"
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 7
    assert np.min(hist.data) == 0

    hist = histogram(locdata_blobs_2d, n_bins=10, bin_range=((500, 1000), (500, 1000)))
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 5
    assert hist.data.shape == (10, 10)

    new_bins = Bins(n_bins=10, bin_range=((500, 1000), (500, 1000)))
    hist = histogram(locdata_blobs_2d, bins=new_bins)
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 5
    assert hist.data.shape == (10, 10)

    new_bins = Bins(n_bins=10, bin_range=((500, 1000), (500, 1000)))
    hist = histogram(locdata_blobs_2d, bin_edges=new_bins.bin_edges)
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 5
    assert hist.data.shape == (10, 10)

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties="position_x")
    assert hist.labels == ["position_x", "counts"]
    assert hist.data.shape == (89,)

    with pytest.raises(ValueError):
        histogram(
            locdata_blobs_2d,
            bin_size=10,
            loc_properties="position_x",
            bin_range=((500, 1000), (500, 1000)),
        )

    hist = histogram(locdata_blobs_2d, bin_size=10, loc_properties=["position_x"])
    assert hist.labels == ["position_x", "counts"]
    assert hist.data.shape == (89,)

    hist = histogram(
        locdata_blobs_2d, bin_size=10, loc_properties=["position_x", "position_y"]
    )
    assert hist.labels == ["position_x", "position_y", "counts"]
    assert hist.data.shape == (89, 55)

    hist = histogram(
        locdata_blobs_2d, bin_size=10, loc_properties=["position_x", "cluster_label"]
    )
    assert hist.labels == ["position_x", "cluster_label", "counts"]
    assert hist.data.shape == (89, 1)

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties="position_z")

    with pytest.raises(ValueError):
        histogram(locdata_blobs_2d, bin_size=10, loc_properties=["position_z"])

    with pytest.raises(ValueError):
        histogram(
            locdata_blobs_2d, bin_size=10, loc_properties=["position_x", "position_z"]
        )

    hist = histogram(
        locdata_blobs_2d,
        bin_edges=((500, 600, 700, 800, 900, 1000), (500, 600, 700, 800, 900, 1000)),
    )
    assert hist.data.ndim == 2
    assert np.max(hist.data) == 7

    hist = histogram(locdata_blobs_2d, bin_size=10, other_property="position_y")
    assert hist.labels == ["position_x", "position_y", "position_y"]
    assert hist.data.shape == (89, 55)
    assert np.isnan(hist.data).any()
    assert np.nanmax(hist.data) == 1001
    assert np.nanmin(hist.data) == 465


def test_histogram_1d(locdata_1d):
    hist = histogram(locdata_1d, n_bins=10)
    assert hist.labels == ["position_x", "counts"]
    assert hist.data.dtype == "float64"
    assert hist.data.ndim == 1
    assert np.max(hist.data) == 2
    assert hist.data.shape == (10,)

    hist = histogram(locdata_1d, n_bins=5, bin_range=(5, 10))
    assert np.max(hist.data) == 1
    assert hist.data.shape == (5,)

    hist = histogram(locdata_1d, bin_edges=(5, 6, 7, 8, 9, 10))
    assert hist.data.shape == (5,)

    hist = histogram(locdata_1d, n_bins=10, other_property="intensity")
    assert hist.labels == ["position_x", "intensity"]
    assert hist.data.shape == (10,)
    assert np.isnan(hist.data).any()
    assert np.nanmax(hist.data) == 125


def test_histogram_3d(locdata_blobs_3d):
    hist = histogram(locdata_blobs_3d, n_bins=10)
    assert hist.labels == ["position_x", "position_y", "position_z", "counts"]
    assert hist.data.dtype == "float64"
    assert hist.data.ndim == 3
    assert np.max(hist.data) == 6
    assert hist.data.shape == (10, 10, 10)

    hist = histogram(
        locdata_blobs_3d, n_bins=10, bin_range=((500, 1000), (500, 1000), (500, 1000))
    )
    assert np.max(hist.data) == 4
    assert hist.data.shape == (10, 10, 10)

    hist = histogram(
        locdata_blobs_3d,
        bin_edges=(
            (500, 600, 700, 800, 900, 1000),
            (500, 600, 700, 800, 900, 1000),
            (500, 600, 700, 800, 900, 1000),
        ),
    )
    assert hist.data.shape == (5, 5, 5)

    hist = histogram(locdata_blobs_3d, n_bins=10, other_property="position_y")
    assert hist.labels == ["position_x", "position_y", "position_z", "position_y"]
    assert hist.data.shape == (10, 10, 10)
    assert np.isnan(hist.data).any()
    assert np.nanmax(hist.data) == 787


def test_histogram_empty(locdata_empty):
    with pytest.raises(TypeError):
        histogram(locdata_empty, n_bins=10)


def test_histogram_single_value(locdata_single_localization_3d):
    hist = histogram(locdata_single_localization_3d, n_bins=3)
    assert hist.data.shape == (3, 3, 3)
    assert np.array_equal(hist.bins.bin_range, [[1, 2], [1, 2], [1, 2]])

    hist = histogram(locdata_single_localization_3d, bin_size=0.2)
    assert hist.data.shape == (5, 5, 5)
    assert np.array_equal(
        hist.bins.bin_range,
        [[1, pytest.approx(2)], [1, pytest.approx(2)], [1, pytest.approx(2)]],
    )

    hist = histogram(locdata_single_localization_3d, bin_size=2)
    assert hist.data.shape == (1, 1, 1)
    assert np.array_equal(hist.bins.bin_range, [[1, 2], [1, 2], [1, 2]])


def test_histogram_2d_negative_values(locdata_2d_negative):
    hist = histogram(locdata_2d_negative, n_bins=10)
    assert hist.labels == ["position_x", "position_y", "counts"]
    assert hist.data.shape == (10, 10)


def test__accumulate_1d():
    bins = Bins(n_bins=5, bin_range=(0, 10))
    n_bins = bins.n_bins[0] + 2  # including over- and underflow bins

    # test 0 - no data
    data = np.array([])
    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data,
        bin_edges=bins.bin_edges[0],
        return_data=True,
        return_counts=True,
    )
    # print(f"{bin_indices=}")
    # print(f"{data_indices=}")
    # print(f"{collection=}")
    assert np.array_equal(bin_indices, [])
    assert data_indices == collection == []
    assert np.array_equal(counts, [])

    # test 1 - single data point
    data = np.array([9])
    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data, bin_edges=bins.bin_edges[0], return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [5])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(data_indices, [np.array([0])])
        ]
    )
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(collection, [np.array([9])])
        ]
    )
    assert np.array_equal(counts, [1])

    # test 2 - return_data=False
    data = np.array([1, -1, 11, 0.1, 6, 2, 1, 2.0])

    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data, bin_edges=bins.bin_edges[0]
    )
    print(data_indices)
    assert np.array_equal(bin_indices, [0, 1, 2, 4, 6])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                data_indices,
                [
                    np.array([1]),
                    np.array([0, 3, 6]),
                    np.array([5, 7]),
                    np.array([4]),
                    np.array([2]),
                ],
            )
        ]
    )
    assert collection is None
    assert counts is None

    # test 3
    data = np.array([1, -1, 11, 0.1, 6, 2, 1, 2.0])

    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data, bin_edges=bins.bin_edges[0], return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [0, 1, 2, 4, 6])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                data_indices,
                [
                    np.array([1]),
                    np.array([0, 3, 6]),
                    np.array([5, 7]),
                    np.array([4]),
                    np.array([2]),
                ],
            )
        ]
    )
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                collection,
                [
                    np.array([-1]),
                    np.array([1, 0.1, 1]),
                    np.array([2, 2]),
                    np.array([6]),
                    np.array([11]),
                ],
            )
        ]
    )
    assert np.array_equal(counts, [1, 3, 2, 1, 1])

    # test 4 - large data
    n_samples = 1_000
    data = np.random.default_rng().normal(loc=5, scale=5, size=n_samples)

    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data, bin_edges=bins.bin_edges[0], return_data=True, return_counts=True
    )
    assert (
        len(bin_indices)
        == len(data_indices)
        == len(collection)
        == len(counts)
        == n_bins
    )

    # test 5 - Two dimensions
    data = np.array(
        [(1, 2), (-1, 5), (11, -1), (0.1, 11), (6, 5), (2, 5), (1, 2.1), (2.0, 2)]
    )

    bin_indices, data_indices, collection, counts = _accumulate_1d(
        data, bin_edges=bins.bin_edges[0], return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [0, 1, 2, 4, 6])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                data_indices,
                [
                    np.array([1]),
                    np.array([0, 3, 6]),
                    np.array([5, 7]),
                    np.array([4]),
                    np.array([2]),
                ],
            )
        ]
    )
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                collection,
                [
                    np.array([(-1, 5)]),
                    np.array([(1, 2), (0.1, 11), (1, 2.1)]),
                    np.array([(2, 5), (2, 2)]),
                    np.array([(6, 5)]),
                    np.array([(11, -1)]),
                ],
            )
        ]
    )
    assert np.array_equal(counts, [1, 3, 2, 1, 1])


def test__accumulate_2d():
    bins = Bins(n_bins=5, bin_range=((0, 10), (0, 10)))
    n_bins = bins.n_bins[0] + 2  # including over- and underflow bins

    # test 0 - no data
    data = []

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data=data, bin_edges=bins.bin_edges, return_data=True, return_counts=True
    )
    # print(f"{bin_indices=}")
    # print(f"{data_indices=}")
    # print(f"{collection=}")
    assert np.array_equal(bin_indices, [])
    assert data_indices == collection == []
    assert np.array_equal(counts, [])

    # test 1 - single data point
    data = [(1, 1)]

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data=data, bin_edges=bins.bin_edges, return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [[1, 1]])
    assert np.array_equal(data_indices, [[0]])
    assert np.array_equal(collection, [[(1, 1)]])
    assert np.array_equal(counts, [1])

    # test 2 - return_data=False
    data = np.array([(1, 2), (1, 5), (4, 5), (1, 2)])

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data=data, bin_edges=bins.bin_edges
    )
    assert np.array_equal(bin_indices, [[1, 2], [1, 3], [3, 3]])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(data_indices, [[0, 3], [1], [2]])
        ]
    )
    assert collection is None
    assert counts is None

    # test 3 - including over- underflow bins
    data = np.array([(1, 2), (-1, 5), (-1, -1), (2, 11), (1, 5), (4, 5), (1.1, 2.1)])

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data=data, bin_edges=bins.bin_edges, return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [[0, 0], [0, 3], [1, 2], [1, 3], [2, 6], [3, 3]])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(data_indices, [[2], [1], [0, 6], [4], [3]])
        ]
    )
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                collection,
                [
                    [[-1, -1]],
                    [[-1, 5]],
                    [[1, 2], [1.1, 2.1]],
                    [[1, 5]],
                    [[2, 11]],
                    [[4, 5]],
                ],
            )
        ]
    )
    assert np.array_equal(counts, [1, 1, 2, 1, 1, 1])

    # test 4 - large data
    n_samples = 1_000
    data = np.random.default_rng().normal(loc=5, scale=5, size=(n_samples, 2))

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data, bin_edges=bins.bin_edges, return_data=True, return_counts=True
    )
    assert (
        len(bin_indices)
        == len(data_indices)
        == len(collection)
        == len(counts)
        == n_bins**2
    )

    # test 5 - Three dimensions
    data = np.array(
        [
            (1, 2, 1),
            (-1, 5, 1),
            (-1, -1, 1),
            (2, 11, 1),
            (1, 5, 1),
            (4, 5, 1),
            (1.1, 2.1, 1),
        ]
    )

    bin_indices, data_indices, collection, counts = _accumulate_2d(
        data, bin_edges=bins.bin_edges, return_data=True, return_counts=True
    )
    assert np.array_equal(bin_indices, [[0, 0], [0, 3], [1, 2], [1, 3], [2, 6], [3, 3]])
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(data_indices, [[2], [1], [0, 6], [4], [3]])
        ]
    )
    assert np.all(
        [
            np.array_equal(result, expected)
            for result, expected in zip(
                collection,
                [
                    [[-1, -1, 1]],
                    [[-1, 5, 1]],
                    [[1, 2, 1], [1.1, 2.1, 1]],
                    [[1, 5, 1]],
                    [[2, 11, 1]],
                    [[4, 5, 1]],
                ],
            )
        ]
    )
    assert np.array_equal(counts, [1, 1, 2, 1, 1, 1])



import matplotlib.pyplot as plt  # this import is needed for interactive tests

from locan import scatter_3d_mpl


def test_scatter_3d_mpl(locdata_3d):
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    scatter_3d_mpl(locdata_3d, ax=ax, text_kwargs=dict(color="r"), color="r")
    # plt.show()

    plt.close("all")


def test_scatter_3d_mpl_empty(locdata_empty):
    fig = plt.figure()
    ax = fig.add_subplot(projection="3d")
    scatter_3d_mpl(locdata_empty, ax=ax)
    # plt.show()

    plt.close("all")


def test_scatter_3d_mpl_single(locdata_single_localization, caplog):
    fig = plt.figure()
    fig.add_subplot(projection="3d")
    scatter_3d_mpl(locdata_single_localization)
    log_msg = (
        "locan.visualize.render_mpl.render3d",
        30,
        "Locdata carries a single localization.",
    )
    assert log_msg in caplog.record_tuples
    # plt.show()

    plt.close("all")



"""

File input/output for localization data in SMLM files.

File specifications are provided at https://github.com/imodpasteur/smlm-file-format/blob/master/specification.md.

Code is adapted from https://github.com/imodpasteur/smlm-file-format/blob/master/implementations/Python/smlm_file.py.
(MIT license)
"""

from __future__ import annotations

import json
import logging
import os
import time
import zipfile
from typing import IO, Any, cast

import numpy as np
import numpy.typing as npt
import pandas as pd
from google.protobuf import json_format

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata import manifest_pb2
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)
from locan.utils.format import _time_string

__all__: list[str] = [
    "manifest_format_from_locdata",
    "manifest_file_info_from_locdata",
    "manifest_from_locdata",
    "save_SMLM",
    "load_SMLM_manifest",
    "load_SMLM_header",
    "load_SMLM_file",
]

logger = logging.getLogger(__name__)


dtype2length = {
    "int8": 1,
    "uint8": 1,
    "int16": 2,
    "uint16": 2,
    "int32": 4,
    "uint32": 4,
    "int64": 8,
    "uint64": 8,
    "float32": 4,
    "float64": 8,
}


def manifest_format_from_locdata(
    locdata: LocData,
) -> locan.locan_io.locdata.manifest_pb2.Format:
    """
    Prepare a manifest["format"] protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.Format
        The manifest format
    """
    format = manifest_pb2.Format()

    # required by
    format.name = "smlm-table(binary)-0"
    format.type = manifest_pb2.Type.TABLE  # type: ignore
    format.mode = manifest_pb2.Mode.BINARY  # type: ignore
    format.columns = len(locdata.data.columns)
    format.headers.extend(locdata.data.columns)
    format.dtype.extend(
        [manifest_pb2.Dtype.Value(str(dt).upper()) for dt in locdata.data.dtypes.values]  # type: ignore
    )
    format.shape.extend([1] * len(locdata.data.columns))
    format.units.extend([""] * len(locdata.data.columns))  # todo: add correct units

    # optional by specifications
    format.description = "localization table generated from locan.LocData"

    return format


def manifest_file_info_from_locdata(
    locdata: LocData,
) -> locan.locan_io.locdata.manifest_pb2.FileInfo:
    """
    Prepare a manifest["file_info"] protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.FileInfo
        The manifest file information
    """
    file_info = manifest_pb2.FileInfo()

    # required by specifications
    file_info.name = "table-0.bin"
    file_info.type = manifest_pb2.TABLE
    file_info.format = "smlm-table(binary)-0"
    file_info.channel = "default"
    file_info.rows = len(locdata)
    for column in locdata.data.columns:
        file_info.offset[column] = 0

    # optional by specifications
    for column in locdata.data.columns:
        file_info.min[column] = locdata.data[column].min()
    for column in locdata.data.columns:
        file_info.max[column] = locdata.data[column].max()
    # file_info.exposure = 20

    return file_info


def manifest_from_locdata(
    locdata: LocData, return_json_string: bool = False
) -> locan.locan_io.locdata.manifest_pb2.Manifest | str:
    """
    Prepare a manifest protobuf from locdata.
    The manifest holds metadata for smlm files.

    Parameters
    ----------
    locdata : LocData
        The LocData object.
    return_json_string : bool
        Flag for returning json string

    Returns
    -------
    locan.locan_io.locdata.manifest_pb2.Manifest | str
        The manifest
    """
    manifest = manifest_pb2.Manifest()
    format = manifest_format_from_locdata(locdata)
    file_info = manifest_file_info_from_locdata(locdata)

    # required by specifications
    manifest.format_version = "0.2"
    manifest.formats["smlm-table(binary)-0"].CopyFrom(format)
    manifest.files.append(file_info)

    # recommended by specifications
    manifest.name = locdata.meta.identifier
    manifest.description = locdata.meta.comment
    manifest.tags.append("SMLM")
    manifest.thumbnail = ""
    manifest.sample = ""
    manifest.labeling = ""
    manifest.date = _time_string(time.time())

    # optional by specifications
    manifest.author = ""
    manifest.citation = ""
    manifest.email = ""
    manifest.locdata_meta.CopyFrom(locdata.meta)

    if return_json_string:
        json_string = json_format.MessageToJson(
            manifest,
            preserving_proto_field_name=True,
            always_print_fields_with_no_presence=False,
        )
        json_string = _change_upper_to_lower_keys(json_string)
        return json_string
    else:
        return manifest


def _change_upper_to_lower_keys(json_string: str) -> str:
    """Switch selected key words in json_string from upper to lower letters."""
    json_string = json_string.replace("BINARY", "binary")
    json_string = json_string.replace("TEXT", "text")
    json_string = json_string.replace("TABLE", "table")
    json_string = json_string.replace("IMAGE", "image")
    json_string = json_string.replace("INT", "int")
    json_string = json_string.replace("UINT", "uint")
    json_string = json_string.replace("FLOAT", "float")
    return json_string


def save_SMLM(
    locdata: LocData,
    path: str | os.PathLike[Any] | IO[Any],
    manifest: locan.locan_io.locdata.manifest_pb2.Manifest | None = None,
) -> None:
    """
    Save LocData attributes in a SMLM single-molecule
    localization (zip) file with manifest.json (version 0.2).

    In the smlm file format we store metadata as a human-
    readable manifest. The data is stored as byte string.

    Note
    ----
    Only selected LocData attributes are saved.
    These are: 'data', 'columns', 'meta'.

    Parameters
    ----------
    locdata
        The LocData object to be saved.
    path
        File path including file name to save to.
    manifest
        Protobuf with manifest to use instead of an autogenerated manifest.
    """
    if manifest is None:
        manifest_json = manifest_from_locdata(locdata, return_json_string=True)
    elif isinstance(manifest, str):
        manifest_json = manifest
    elif isinstance(manifest, manifest_pb2.Manifest):
        manifest_json = json_format.MessageToJson(
            manifest,
            preserving_proto_field_name=True,
            always_print_fields_with_no_presence=False,
        )
        manifest_json = _change_upper_to_lower_keys(manifest_json)
    else:
        raise TypeError("Type of manifest is not correct.")

    byte_string = b"".join(
        [
            locdata.data[column][i].tobytes()
            for i in range(len(locdata))
            for column in locdata.data.columns
        ]
    )

    # save zip file
    with zipfile.ZipFile(
        path, "w", compression=zipfile.ZIP_DEFLATED, allowZip64=True
    ) as zf:
        # write manifest
        zf.writestr("manifest.json", cast(str, manifest_json))
        # write files
        zf.writestr("table-0.bin", byte_string)


def load_SMLM_manifest(
    path: str | os.PathLike[str] | IO[Any],
) -> dict[str, Any]:
    """
    Read manifest.json (version 0.2) from a SMLM single-molecule localization
    (zip) file.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    dict[str, Any]
        manifest in json format
    """
    zf = zipfile.ZipFile(path, "r")
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception("invalid file: no manifest.json found in the smlm file.")
    manifest: dict[str, Any] = json.loads(zf.read("manifest.json"))
    if manifest["format_version"] != "0.2":
        raise NotImplementedError("Not implemented for manifest version unlike 0.2.")
    return manifest


def load_SMLM_header(
    path: str | os.PathLike[str] | IO[Any],
) -> list[str]:
    """
    Read header (manifest) from a SMLM single-molecule localization file and
    identify column names.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    list[str]
        A list of dataset property keys as derived from the identifiers.
    """
    zf = zipfile.ZipFile(path, "r")
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception("invalid file: no manifest.json found in the smlm file.")
    manifest = json.loads(zf.read("manifest.json"))
    if manifest["format_version"] != "0.2":
        raise NotImplementedError("Not implemented for manifest version unlike 0.2.")

    locdata_columns_list = []
    for file_info in manifest["files"]:
        if file_info["type"] != "table":
            logger.info("ignore file with type: %s", file_info["type"])
        else:
            name = file_info["name"]
            logger.info(f"loading table {name}....")
            format_key = file_info["format"]
            file_format = manifest["formats"][format_key]
            if name not in file_names:
                logger.error("ERROR: Did not find %s in zip file", file_info["name"])
            else:
                headers = file_format["headers"]

                column_keys = convert_property_names(
                    properties=headers, property_mapping=locan.constants.SMLM_KEYS
                )
                locdata_columns_list.append(column_keys)

    if len(locdata_columns_list) == 1:
        return locdata_columns_list[0]
    else:
        return locdata_columns_list  # type: ignore


def load_SMLM_file(
    path: str | os.PathLike[Any] | IO[Any],
    nrows: int | None = None,
    convert: bool = True,
) -> LocData | list[LocData]:
    """
    Load data from a SMLM single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file.
        None means that all available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData | list[LocData]
        A new instance of LocData with all localizations. Returns a list of
        LocData if multiple tables are found.
    """
    zf = zipfile.ZipFile(path, "r")
    file_names = zf.namelist()
    if "manifest.json" not in file_names:
        raise Exception("invalid file: no manifest.json found in the smlm file.")
    manifest = json.loads(zf.read("manifest.json"))
    if manifest["format_version"] != "0.2":
        raise NotImplementedError("Not implemented for manifest version unlike 0.2.")

    locdatas = []
    for file_info in manifest["files"]:
        if file_info["type"] != "table":
            logger.info("ignore file with type: %s", file_info["type"])
        else:
            name = file_info["name"]
            logger.debug(f"start loading {name} ...")
            format_key = file_info["format"]
            file_format = manifest["formats"][format_key]
            if file_format["mode"] != "binary":
                raise Exception(f"format mode {file_format['mode']} not supported.")
            else:
                try:
                    table_file = zf.read(file_info["name"])
                except KeyError:
                    logger.error(
                        "ERROR: Did not find %s in zip file", file_info["name"]
                    )
                    continue
                else:
                    logger.debug(f"loading {len(table_file)} bytes")
                    headers = file_format["headers"]
                    dtype = file_format["dtype"]
                    shape = file_format["shape"]
                    cols = len(headers)
                    rows = nrows if nrows is not None else int(file_info["rows"])

                    logger.debug("columns: %s", headers)
                    logger.debug("rows: %s, columns: %s", rows, cols)
                    if not len(headers) == len(dtype) == len(shape):
                        raise ValueError(
                            "headers, dtype, and shape have not the same length."
                        )

                    rowLen = sum(
                        dtype2length[dtype[i]] for i, header in enumerate(headers)
                    )
                    tableDict: dict[str, npt.NDArray[Any]] = {}
                    byteOffset = 0
                    for i, header in enumerate(headers):
                        tableDict[header] = np.ndarray(
                            (rows,),
                            buffer=table_file,
                            dtype=dtype[i],
                            offset=byteOffset,
                            order="C",
                            strides=(rowLen,),
                        )
                        byteOffset += dtype2length[dtype[i]]
                    logger.debug(f"finished loading {name}")

                    dataframe = pd.DataFrame.from_dict(tableDict)

                    column_keys = {}
                    for header in headers:
                        if header in locan.constants.SMLM_KEYS:
                            column_keys[header] = locan.constants.SMLM_KEYS[header]
                        elif header in locan.constants.RAPIDSTORM_KEYS:
                            column_keys[header] = locan.constants.RAPIDSTORM_KEYS[
                                header
                            ]
                        else:
                            logger.warning(
                                f"Column {header} is not a Locan property standard."
                            )
                            column_keys[header] = header
                    dataframe.rename(columns=column_keys, inplace=True)

                    if convert:
                        dataframe = convert_property_types(
                            dataframe, types=locan.constants.PROPERTY_KEYS
                        )

                    locdata = LocData.from_dataframe(dataframe=dataframe)

                    # todo: check on taking correct metadata
                    locdata.meta.source = metadata_pb2.EXPERIMENT
                    locdata.meta.state = metadata_pb2.RAW
                    locdata.meta.file.type = metadata_pb2.SMLM
                    locdata.meta.file.path = str(path)

                    for property_ in sorted(
                        list(
                            set(dataframe.columns).intersection(
                                {"position_x", "position_y", "position_z"}
                            )
                        )
                    ):
                        locdata.meta.localization_properties.add(
                            name=property_, unit="nm", type="float"
                        )

                    del locdata.meta.history[:]
                    locdata.meta.history.add(
                        name="load_SMLM_file", parameter=f"path={path}, nrows={nrows}"
                    )

                    locdatas.append(locdata)

    if len(locdatas) == 1:
        return locdatas[0]
    else:
        return locdatas



"""
Region objects for localization data.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   region
   region_utils

"""

from __future__ import annotations

from locan.data.regions.region import *
from locan.data.regions.region_utils import *

from . import region, region_utils

__all__: list[str] = []
__all__.extend(region.__all__)
__all__.extend(region_utils.__all__)



"""

File input/output for localization data in Thunderstorm files.

"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead, SupportsReadline, SupportsWrite

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
    open_path_or_file_like,
)

__all__: list[str] = [
    "load_thunderstorm_header",
    "load_thunderstorm_file",
    "save_thunderstorm_csv",
]

logger = logging.getLogger(__name__)


def _read_thunderstorm_header(file: SupportsReadline[Any]) -> list[str]:
    """
    Read csv header from a Thunderstorm single-molecule localization file and
    identify column names.

    Parameters
    ----------
    file : SupportsReadline
        Thunderstorm file to read.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the Thunderstorm
        identifiers.
    """
    # read csv header
    header = file.readline().split("\n")[0]

    # list identifiers
    identifiers = [x.strip('"') for x in header.split(",")]

    column_keys = convert_property_names(
        properties=identifiers, property_mapping=locan.constants.THUNDERSTORM_KEYS
    )
    return column_keys


def load_thunderstorm_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> list[str]:
    """
    Load csv header from a Thunderstorm single-molecule localization file and
    identify column names.

    Parameters
    ----------
    path
        File path for a file to load.

    Returns
    -------
    list[str]
        A list of valid dataset property keys as derived from the
        identifiers.
    """
    # read csv header
    with open_path_or_file_like(path) as file:
        return _read_thunderstorm_header(file)


def load_thunderstorm_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
    **kwargs: Any,
) -> LocData:
    """
    Load data from a Thunderstorm single-molecule localization file.

    Parameters
    ----------
    path
        File path for a file to load.
    nrows
        The number of localizations to load from file. None means that all
        available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.
    kwargs
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with open_path_or_file_like(path) as file:
        columns = _read_thunderstorm_header(file)
        dataframe = pd.read_csv(
            file, sep=",", skiprows=0, nrows=nrows, names=columns, **kwargs
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.THUNDERSTORM
    dat.meta.file.path = str(path)

    for property_ in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        dat.meta.localization_properties.add(name=property_, unit="nm", type="float")

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_thundestorm_file", parameter=f"path={str(path)}, nrows={nrows}"
    )

    return dat


def save_thunderstorm_csv(
    locdata: LocData,
    path: str | os.PathLike[Any] | SupportsWrite[Any],
) -> None:
    """
    Save LocData attributes Thunderstorm-readable csv-file.

    In the Thunderstorm csv-file file format we store only localization data
    with Thunderstorm-readable column names.

    Parameters
    ----------
    locdata
        The LocData object to be saved.
    path
        File path including file name to save to.
    """
    # get data from locdata object
    dataframe = locdata.data

    # create reverse mapping to Thunderstorm columns
    inv_map = {v: k for k, v in locan.constants.THUNDERSTORM_KEYS.items()}

    # rename columns
    dataframe = dataframe.rename(index=str, columns=inv_map, inplace=False)

    # write to csv
    dataframe.to_csv(path, float_format="%.10g", index=False)  # type: ignore[arg-type]



"""
File manager

Identify, match, and group files to be batch-processed.
The class Files is a wrapper for a pandas.DataFrame with selected methods to
identify, match, and group file paths.
"""

from __future__ import annotations

import functools
import logging
import os
import re
import sys
from collections.abc import Iterable, Sequence
from pathlib import Path
from typing import Any

from pandas import DataFrame, Series

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import pandas as pd

from locan.locan_io.utilities import find_file_upstream

__all__: list[str] = ["Files"]

logger = logging.getLogger(__name__)


class Files:
    """
    Wrapper for a pandas.DataFrame with selected methods to
    identify, match, and group file paths.

    Note
    ------
    Iteration and indexing is implemented in a way that integer indexing
    or iterating over the Files instance returns
    a single row (as Series or namedtuple).
    Slice indexing returns a new Files instance with selected rows.

    Parameters
    ----------
    df : pd.DataFrame | dict[str, str] | None
        file names
    directory : str | os.PathLike[Any] | None
        base directory
    exists : bool
        raise `FileExistsError` if file in `df` does not exist
    column : str
        key/column in `df` from which to take a file list

    Attributes
    ----------
    df : pd.DataFrame
        dataframe carrying file paths
    directory: Path
        base directory
    """

    def __init__(
        self,
        df: pd.DataFrame | dict[str, str] | None = None,
        directory: str | os.PathLike[Any] | None = None,
        exists: bool = True,
        column: str = "file_path",
    ) -> None:
        if directory is None:
            self.directory = None
        elif isinstance(directory, (str, os.PathLike)):
            self.directory = Path(directory)
            if not self.directory.exists():
                raise ValueError(f"The directory {directory} does not exist.")
        else:
            raise TypeError("The given directory is not a valid type.")

        if df is None:
            self.df = pd.DataFrame(columns=[column])
        elif isinstance(df, (pd.DataFrame, pd.Series)):
            if hasattr(df, column):
                self.df = df
            else:
                raise AttributeError(f"Dataframe must have column {column}.")
        else:
            try:
                self.df = pd.DataFrame(data=df[column], columns=[column])
            except AttributeError as err:
                raise AttributeError(f"Dataframe must have column {column}.") from err

        if (
            not self.df.empty
            and not self.df[column].apply(lambda x: isinstance(x, Path)).all()
        ):
            self.df[column] = self.df[column].astype("string").map(Path)

        if exists:
            mask = ~self.df[column].apply(lambda x: x.exists())
            non_existing = self.df[column][mask]
            if any(mask):
                raise FileExistsError(
                    f"The following files do not exist: {non_existing}"
                )

    def __iter__(self) -> Iterable[tuple[Any, ...]]:
        return_value: Iterable[tuple[Any, ...]] = self.df.itertuples(name="Files")
        return return_value

    def __getitem__(self, index: Any) -> Series[Any] | DataFrame | Files:
        if isinstance(index, int):
            return self.df.iloc[index]
        else:
            return Files(directory=self.directory, df=self.df.iloc[index])

    @classmethod
    def concatenate(
        cls,
        files: Iterable[Files] | None = None,
        directory: str | os.PathLike[Any] | None = None,
        exists: bool = True,
    ) -> Files:
        """
        Concatenate the file lists from multiple File instances
        and set the base directory without further action.

        Parameters
        ----------
        files
            sequence with File instances
        directory
            new base directory
        exists
            raise `FileExistsError` if file in files does not exist

        Returns
        -------
        Files
        """
        if files is None:
            return cls()
        else:
            df = pd.concat([files_.df for files_ in files], join="inner")
            df = df.drop_duplicates().reset_index(drop=True)
            column = df.columns[0]
            return cls(df=df, directory=directory, exists=exists, column=column)

    @classmethod
    def from_path(
        cls,
        files: Sequence[str | os.PathLike[Any]] | str | os.PathLike[Any] | None = None,
        directory: str | os.PathLike[Any] | None = None,
        column: str = "file_path",
    ) -> Files:
        """
        Instantiate `Files` from a collection of file paths.

        Parameters
        ----------
        files
            sequence with File instances
        directory
            new base directory
        column
            Name of column in `Files.df` carrying these files

        Returns
        -------
        Files
        """
        df: pd.DataFrame | dict[str, str] | None
        if isinstance(files, (str, os.PathLike)):
            df = pd.DataFrame(data=[files], columns=[column])
        elif isinstance(files, Iterable):
            df = pd.DataFrame(data=files, columns=[column])
        else:
            df = files
        return cls(directory=directory, df=df, column=column)

    @classmethod
    def from_glob(
        cls,
        directory: str | os.PathLike[Any] | None = None,
        pattern: str = "*.txt",
        regex: str | None = None,
        column: str = "file_path",
    ) -> Files:
        """
        Instantiate `Files` from a search with glob and/or regex patterns.

        Parameters
        ----------
        pattern
            glob pattern passed to :func:`Path.glob`
        regex
            regex pattern passed to :func:`re.search` and applied in addition
            to glob pattern
        directory
            new base directory in which to search
        column
            Name of column in `Files.df` carrying these files

        Returns
        -------
        Files
        """
        if directory is None:
            directory = Path().cwd()
        else:
            directory = Path(directory)

        files = directory.glob(pattern)

        if regex is None:
            files = list(files)  # type: ignore[assignment]
        else:
            regex_ = re.compile(regex)
            files = [file_ for file_ in files if regex_.search(str(file_)) is not None]  # type: ignore[assignment]

        df = pd.DataFrame(data=files, columns=[column])  # type: ignore[arg-type]
        return cls(directory=directory, df=df, column=column)

    def add_glob(
        self,
        pattern: str | None = "*.txt",
        regex: str | None = None,
        column: str = "other_file_path",
    ) -> Self:
        """
        Search for file paths using glob and/or regex pattern in base directory
        and provide files in new `column`.

        A logging.warning is given if the number of found files and those
        in `self.df` are different.

        Parameters
        ----------
        pattern
            glob pattern passed to :func:`Path.glob`
        regex
            regex pattern passed to :func:`re.search` and applied in addition
            to glob pattern
        column : str
            Name of column in `Files.df` carrying these files

        Returns
        -------
        Self
        """
        files = self.directory.glob(pattern)  # type: ignore

        if regex is None:
            files = list(files)
        else:
            regex_ = re.compile(regex)
            files = [file_ for file_ in files if regex_.search(str(file_)) is not None]

        self.df[column] = pd.Series(files)
        if len(self.df) != self.df[column].count():
            logger.warning("Not all files are matched.")
        return self

    def exclude(
        self,
        stoplist: Files | Iterable[bool | str | os.PathLike[Any]] | None = None,
        column: str = "file_path",
        column_stoplist: str = "file_path",
    ) -> Self:
        """
        Exclude files in `self.df.column` according to stoplist.

        Parameters
        ----------
        stoplist
            Files to be excluded
        column
            key/column in `df` from which to exclude files
        column_stoplist
            key/column in `stoplist` from which to take files

        Returns
        -------
        Self
        """
        if stoplist is None or bool(stoplist) is False:
            pass
        elif not isinstance(stoplist, Files) and all(
            isinstance(item, bool) for item in stoplist
        ):
            selection = [not item_ for item_ in stoplist]
            self.df = self.df[selection]
        else:
            if isinstance(stoplist, Files):
                stoplist = stoplist.df[column_stoplist].astype("string")
            else:
                try:
                    stoplist = [str(item_[column_stoplist]) for item_ in stoplist]  # type: ignore[index]
                except TypeError:
                    stoplist = [str(item_) for item_ in stoplist]

            if len(stoplist) != 0:  # type: ignore[arg-type]
                conditions = [
                    self.df[column].astype("string").str.contains(item_, regex=False)  # type: ignore[arg-type]
                    for item_ in stoplist  # type: ignore
                ]
                mask = functools.reduce(lambda x, y: x | y, conditions)
                self.df = self.df[~mask]
        return self

    def match_files(
        self,
        files: pd.Series[Any],
        column: str = "file_path",
        other_column: str = "other_file_path",
    ) -> Self:
        """
        Add files in new column.

        A logging.warning is given if the number of files and those
        in `self.df` are different.

        Parameters
        ----------
        files
            New file list
        column
            Name of column in `Files.df` carrying files to match
        other_column
            Name of new column carrying files

        Returns
        -------
        Self
        """
        self.df[other_column] = files
        if self.df[column].count() != self.df[other_column].count():
            logger.warning("Not all files are matched.")
        return self

    def match_file_upstream(
        self,
        column: str = "file_path",
        pattern: str | None = "*.toml",
        regex: str | None = None,
        directory: str | os.PathLike[Any] | None = None,
        other_column: str = "metadata",
    ) -> Self:
        """
        Find a matching file by applying :func:`locan.find_file_upstream` on
        each file in `self.df[column]`.

        Parameters
        ----------
        column
            Name of column in `Files.df` carrying files to match
        pattern
            glob pattern passed to :func:`Path.glob`
        regex
            regex pattern passed to :func:`re.search` and applied in addition
            to glob pattern
        directory
            top directory in which to search
        other_column
            Name of new column carrying files

        Returns
        -------
        Self
        """
        matched_file = [
            find_file_upstream(
                sub_directory=file_,
                pattern=pattern,
                top_directory=directory,
                regex=regex,
            )
            for file_ in self.df[column]
        ]
        self.df[other_column] = matched_file
        return self

    def print_summary(self) -> None:
        """
        Print summary of Files.

        Returns
        -------
        None
        """
        print(f"Number of files: {len(self.df)}")
        print(f"Base directory: {self.directory}")
        print(f"Columns: {self.df.columns}")
        print(self.df.describe().loc[["count", "unique"]])

    def group_identifiers(self) -> Any:  # todo: fix type
        """
        Get categories defined in self.df.group.

        Returns
        -------
        categories
        """
        return self.df.group.cat.categories

    def grouped(self) -> pd.core.groupby.DataFrameGroupBy[Any]:  # type: ignore[type-arg]
        """
        Get groupby instance based on group_identifiers.

        Returns
        -------
        pandas.core.groupby.DataFrameGroupBy
        """
        return self.df.groupby(by="group", observed=True)

    def set_group_identifier(
        self,
        name: str | None = None,
        pattern: str | None = None,
        glob: str | None = None,
        regex: str | None = None,
        column: str = "file_path",
    ) -> Self:
        """
        Set group_identifier `name` for files in `column` as identified by
        string pattern and/or glob pattern and/or regex
        and keep them in column "group".

        Parameters
        ----------
        name
            new group_identifier
        pattern
            string pattern
        glob
            glob pattern passed to :func:`Path.match`
        regex
            regex pattern
        column
            Name of column in `Files.df` carrying files to match

        Returns
        -------
        Self
        """
        if all(key_ is None for key_ in [pattern, glob, regex]):
            return self

        if pattern is not None:
            self.df["_mask_pattern"] = (
                self.df[column].astype("string").str.contains(pattern, regex=False)
            )

        if glob is not None:
            self.df["_mask_glob"] = self.df[column].apply(lambda x: x.match(glob))

        if regex is not None:
            self.df["_mask_regex"] = (
                self.df[column].astype("string").str.contains(regex, regex=True)
            )

        mask_columns = [
            key_
            for key_ in ["_mask_pattern", "_mask_glob", "_mask_regex"]
            if key_ in self.df.columns
        ]

        if len(mask_columns) == 1:
            self.df["mask"] = self.df[mask_columns]
        else:
            self.df["mask"] = self.df[mask_columns].all(axis=1)

        if "group" not in self.df.columns:
            self.df["group"] = pd.NA
            self.df["group"] = self.df["group"].astype("category")
        else:
            if self.df.loc[self.df["mask"], "group"].notna().any():
                logger.warning(f"Previously defined groups are overwritten with {name}")

        if name not in self.df["group"].cat.categories:
            self.df["group"] = self.df["group"].cat.add_categories([name])
        self.df.loc[self.df["mask"], "group"] = name
        self.df = self.df.drop(columns=mask_columns + ["mask"])

        if self.df.group.isna().any():
            logger.info("Some group identifiers are still NAN.")
        return self



import matplotlib.pyplot as plt  # needed for visual inspection  # noqa: F401
import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import SubpixelBias
from locan.analysis.subpixel_bias import _subpixel_bias


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test__subpixel_bias(locdata_simple):
    results = _subpixel_bias(locdata=locdata_simple, pixel_size=2)
    assert np.array_equal(results.columns, ["position_x_modulo", "position_y_modulo"])
    assert np.array_equal(results.position_x_modulo.to_numpy(), [0, 0, 1, 0, 1])
    assert np.array_equal(results.position_y_modulo.to_numpy(), [0, 1, 1, 0, 1])


def test_SubpixelBias_empty(caplog):
    SubpixelBias(pixel_size=2).compute(LocData())
    assert caplog.record_tuples == [
        ("locan.analysis.subpixel_bias", 30, "Locdata is empty.")
    ]


def test_SubpixelBias(locdata_simple):
    sb = SubpixelBias(pixel_size=2).compute(locdata_simple)
    assert repr(sb) == "SubpixelBias(pixel_size=2)"
    assert np.array_equal(sb.results.position_x_modulo.to_numpy(), [0, 0, 1, 0, 1])
    assert np.array_equal(sb.results.position_y_modulo.to_numpy(), [0, 1, 1, 0, 1])
    sb.hist()
    # plt.show()

    plt.close("all")


@pytest.mark.visual
def test_DistributionFits(locdata_dSTORM_2d):
    sb = SubpixelBias(pixel_size=130).compute(locdata_dSTORM_2d)
    sb.hist()
    plt.legend()
    plt.show()

    plt.close("all")
